(** @author: Benoît Montagu <benoit.montagu@inria.fr>
    Copyright © Inria 2021 *)

(** This module implements a simple abstract machine that evaluates
   first-order programs with function calls and references. It is a
   variation of the CEK machine. *)

open Ast

type env = (var * v) list
(** Environments associate values to some variables. The environment
   is local to each function. *)

type mu = (loc * v) list
(** Memories associate values to some locations. The memory is global
   to the whole program. *)

type pi = (var option * cmds * env) list
(** Stacks are lists of things to do after a function has returned. They specify:
    - to which variable the returned value must be recorded
    - which is the code to run next, i.e. the remaining of the code of
      the caller that must be executed
      (on real a machine, this would represented by the "return pointer")
    - the environment of the caller that must be restored
*)

type configuration = cmds * env * pi * mu * loc
(** A configuration is the state of the program. It is composed of five elements:
    - the code to execute
    - the local environment
    - the call stack
    - the global memory
    - the state of the memory allocator (here this is the next free address)
*)

(** A helper function that returns the size of the current call stack *)
let stack_size : configuration -> int = function
  | _t, _env, pi, _mu, _loc -> List.length pi

let pp_env fmt env =
  let open Format in
  let open PPrint in
  fprintf fmt "{@[";
  pp_print_list
    ~pp_sep:(fun fmt () -> fprintf fmt ";@ ")
    (fun fmt (x, v) -> fprintf fmt "%a =@ @[%a@]" pp_var x pp_v v)
    fmt env;
  fprintf fmt "@,}@]"

let pp_mu fmt mu =
  let open Format in
  let open PPrint in
  fprintf fmt "{@[";
  pp_print_list
    ~pp_sep:(fun fmt () -> fprintf fmt ";@ ")
    (fun fmt (l, v) -> fprintf fmt "%a ->@ @[%a@]" pp_loc l pp_v v)
    fmt mu;
  fprintf fmt "@,}@]"

let pp_ovar fmt = function
  | None -> Format.pp_print_string fmt "_"
  | Some x -> PPrint.pp_var fmt x

let pp_pi fmt pi =
  let open Format in
  let open PPrint in
  fprintf fmt "@[[";
  pp_print_list
    ~pp_sep:(fun fmt () -> fprintf fmt ";@ ")
    (fun fmt (ox, cmds, env) ->
      fprintf fmt
        "@[<v 1>(var = %a,@ @[<v 2>code =@ %a@],@ @[<v 2>env =@ %a@]@,@])"
        pp_ovar ox pp_cmds cmds pp_env env)
    fmt pi;
  fprintf fmt "@,]@]"

let pp_configuration fmt (cmds, env, pi, mu, loc) =
  let open Format in
  let open PPrint in
  fprintf fmt "{@[<v 0>";
  fprintf fmt "@[<v 2>code =@ %a@];@ " pp_cmds cmds;
  fprintf fmt "@[<v 2>env =@ %a@];@ " pp_env env;
  fprintf fmt "@[<v 2>pi =@ %a@];@ " pp_pi pi;
  fprintf fmt "@[<v 2>mu =@ %a@];@ " pp_mu mu;
  fprintf fmt "@[<v 2>loc =@ %a@];@ " pp_loc loc;
  fprintf fmt "}@]"

exception Error of string
(** The [Error] exception is raised when some illegal state is
   reached. It carries an error message. *)

(** Insertion is an association list whose keys are sorted. *)
let rec insert_sorted x v = function
  | [] -> [ (x, v) ]
  | ((y, _w) as p) :: l as l' ->
      if x < y then (x, v) :: l'
      else if x = y then (x, v) :: l
      else p :: insert_sorted x v l

(** [lookup_env env x] looks up the value associated to the variable
   [x] in the environment [env] *)
let lookup_env (env : env) (x : var) : v =
  try List.assoc x env
  with Not_found -> raise (Error ("Unbound variable " ^ x))

(** [set_env env x v] adds a binding for the variable [x] to the value
   [v] in the environment [env]. If there was already a binding for
   [x] in [env], it is overwritten. *)
let set_env (env : env) (x : var) (v : v) : env = insert_sorted x v env

(** [eval_imm env imm] evaluates the immediate value [imm] in the
   environment [env] *)
let eval_imm (env : env) (imm : imm) : v =
  match imm with
  | Var x -> (
      try List.assoc x env
      with Not_found -> raise (Error ("Unbound variable " ^ x)))
  | Val v -> v

(** [lookup_mem mu loc] looks up the value associated to the location
   [loc] in the memory [mu] *)
let lookup_mem (mu : mu) (loc : loc) : v =
  try List.assoc loc mu
  with Not_found -> raise (Error ("Unbound location " ^ string_of_int loc))

(** [set_mem ~check_presence mu loc v] adds a binding for the location
   [loc] to the value [v] in the memory [mu]. If there was already a
   binding for [l] in [mu], it is overwritten. If [check_presence =
   true], then it is checked that [x] already has a binding in [mu],
   or raises [Error] otherwise. *)
let set_mem ~check_presence (mu : mu) (loc : loc) (v : v) : mu =
  if check_presence && (not @@ List.mem_assoc loc mu) then
    raise
      (Error
         ("Illegal SetRef: this location should be present already "
        ^ string_of_int loc));
  insert_sorted loc v mu

(** [lookup_fun prog f] looks up the function associated to the name
   [f] in the program [prog] *)
let lookup_fun (defs : program) (f : fname) : func =
  try List.assoc f defs
  with Not_found -> raise (Error ("Unbound function " ^ f))

(** [eval_binop op v1 v2] applies the binary operation denoted by [op]
   to the values [v1] and [v2]. Raises [Error] is the operation is not
   defined. *)
let eval_binop op v1 v2 =
  match (op, v1, v2) with
  | AND, Bool b1, Bool b2 -> Bool (b1 && b2)
  | OR, Bool b1, Bool b2 -> Bool (b1 || b2)
  | LE, Int i1, Int i2 -> Bool (i1 <= i2)
  | LT, Int i1, Int i2 -> Bool (i1 < i2)
  | GE, Int i1, Int i2 -> Bool (i1 >= i2)
  | GT, Int i1, Int i2 -> Bool (i1 > i2)
  | NEQ, Int i1, Int i2 -> Bool (i1 <> i2)
  | EQ, Int i1, Int i2 -> Bool (i1 = i2)
  | PLUS, Int i1, Int i2 -> Int (i1 + i2)
  | MINUS, Int i1, Int i2 -> Int (i1 - i2)
  | MULT, Int i1, Int i2 -> Int (i1 * i2)
  | _ -> raise (Error "Illegal binary operation")

(** [is_final conf] returns [Some v] if [conf] in a final
   configuration (i.e. it has an empty call stack) with return value [v],
   or [None] otherwise. *)
let is_final : configuration -> v option = function
  | [ Return (Imm (Val v)) ], _env, [], _mu, _loc -> Some v
  | _ -> None

(** Command that returns the unit value: this is what is produced by
   assignments *)
let return_unit = Return (Imm (Val Unit))

(** [step ~tailcalls program conf] creates a new configuration, that
   is obtained by performing "one step" of evaluation. Raises [Error]
   is some illegal state is reached, or [Done] if [conf] is a final
   state. If [tailcalls = true], then calls in tail positions do not
   consume an extra stack frame.
   Precondition !1: the input configuration is not a final configuration.
   Precondition #2: the input configuration has a non-empty list of commands.
   Precondition #3: the code every function is a non-empty list of commands.
   Ensures: the output configuration has a non-empty list of commands.
*)
let step ~tailcalls (program : program) : configuration -> configuration =
  function
  | [], _, _, _, _ ->
      (* Impossible thanks to precondition #2. *)
      assert false
  | [ Return (Imm (Val _)) ], _env, [], _mu, _loc ->
      (* Impossible thanks to precondition #1. *)
      assert false
  | [ (Return (Imm (Val v)) as retval) ], _env, (ox, cmds', env') :: pi, mu, loc
    ->
      (* If we have obtained a value in a terminal position, and the call
         stack is not empty, then we restore the caller's code and its
         environment, and we associate the returned value to the saved
         variable. *)
      let cmds', env' =
        match ox with
        | Some x -> (return_unit :: cmds', set_env env' x v)
        | None -> (retval :: cmds', env')
      in
      (cmds', env', pi, mu, loc)
  | Return (Imm (Val Unit)) :: cmds, env, pi, mu, loc ->
      (* When we have obtained a unit value in a non-terminal position, then
         we simply execute de rest of the code. *)
      (cmds, env, pi, mu, loc)
  | Return (Imm (Val _)) :: _cmds, _env, _pi, _mu, _loc ->
      (* When we have obtained a value that is not unit, in a non-terminal
         position, then we fail. *)
      raise (Error "Bad sequence: a value was computed, that is not 'unit'.")
  | Return (Imm (Var x)) :: cmds, env, pi, mu, loc ->
      (* To evaluate some variable [x], we just return the value that is
         associated to [x] in the environment [env]. *)
      let vx = lookup_env env x in
      (Return (Imm (Val vx)) :: cmds, env, pi, mu, loc)
  | Return (Binop (op, i1, i2)) :: cmds, env, pi, mu, loc ->
      (* To evaluate a binary operation, we evaluate the operands, and
         return the result of the binary operation *)
      let v1 = eval_imm env i1 and v2 = eval_imm env i2 in
      let v = eval_binop op v1 v2 in
      (Return (Imm (Val v)) :: cmds, env, pi, mu, loc)
  | Return (NewRef i) :: cmds, env, pi, mu, loc ->
      (* To evaluate the creation of a reference, then we evaluate the value
         to which the reference should be initialized, we record that value
         in memory at the next available address, and return that address as
         a result. We also update the state of the memory allocator so that
         it contains the "next" fresh variable. *)
      let vi = eval_imm env i in
      ( Return (Imm (Val (Loc loc))) :: cmds,
        env,
        pi,
        set_mem ~check_presence:false mu loc vi,
        loc + 1 )
  | Return (GetRef x) :: cmds, env, pi, mu, loc -> (
      (* To evaluate the read of a reference, we evaluate the argument
         to a memory location (or fail otherwise), and return the
         value that is held in memory at this location. *)
      let vx = lookup_env env x in
      match vx with
      | Loc locx ->
          (Return (Imm (Val (lookup_mem mu locx))) :: cmds, env, pi, mu, loc)
      | _ -> raise (Error "Illegal GetRef: a location was expected"))
  | Return (SetRef (x, i)) :: cmds, env, pi, mu, loc -> (
      (* To evaluate the write of a reference, we evaluate the first
         argument to a memory location (or fail otherwise), then
         evaluate the second argument, then we update the memory so
         that the memory location points to the computed value, and we
         return the unit value. *)
      let vx = lookup_env env x in
      match vx with
      | Loc locx ->
          let vy = eval_imm env i in
          ( return_unit :: cmds,
            env,
            pi,
            set_mem ~check_presence:true mu locx vy,
            loc )
      | _ -> raise (Error "Illegal SetRef: a location was expected"))
  | If (cond, cmds1, cmds2) :: cmds, env, pi, mu, loc -> (
      (* To evaluate an "if-then-else" statement, we first evaluate
         the branch condition to a boolean value, and then we continue
         with either the "then branch" (when the condition is true) or
         the "else branch" (when the condition is false). *)
      match eval_imm env cond with
      | Bool b -> ((if b then cmds1 else cmds2) @ cmds, env, pi, mu, loc)
      | Unit | Int _ | Loc _ ->
          raise (Error "Illegal 'if': a boolean was expected"))
  | AssignVar (x, Imm i) :: cmds, env, pi, mu, loc ->
      (* To evaluate a variable update with an immediate value, we add a
         binding for this value in the environment, and continue with the
         rest of the code. *)
      (return_unit :: cmds, set_env env x (eval_imm env i), pi, mu, loc)
  | AssignVar (x, Binop (op, i1, i2)) :: cmds, env, pi, mu, loc ->
      (* To evaluate a variable update with a binary operation, we evaluate
         that binary operation to a value, then add a binding for this value
         in the environment, and then continue with the rest of the code. *)
      let v1 = eval_imm env i1 and v2 = eval_imm env i2 in
      let v = eval_binop op v1 v2 in
      (return_unit :: cmds, set_env env x v, pi, mu, loc)
  | AssignVar (x, NewRef i) :: cmds, env, pi, mu, loc ->
      (* To evaluate a variable update with the creation of a reference, we
         evaluate the argument to a value, then we take the next fresh
         memory location, and add a binding to the memory for this location
         pointing to the obtained value. Then, we add a binding for this
         memory location in the environment, and then continue with the rest
         of the code. We also update the next available location, so that it
         corresponds to a fresh location. *)
      let vi = eval_imm env i in
      ( return_unit :: cmds,
        set_env env x (Loc loc),
        pi,
        set_mem ~check_presence:false mu loc vi,
        loc + 1 )
  | AssignVar (x, GetRef y) :: cmds, env, pi, mu, loc -> (
      (* To evaluate a variable update with the read of a reference,
         we evaluate the argument to a location, then we read the
         value in the memory that is recorded at this location, and we
         add a binding in the environment for that value. Then, we
         continue with the rest of the code. *)
      let vy = lookup_env env y in
      match vy with
      | Loc locy ->
          (return_unit :: cmds, set_env env x (lookup_mem mu locy), pi, mu, loc)
      | _ -> raise (Error "Illegal GetRef: a location was expected"))
  | AssignVar (x, SetRef (y, i)) :: cmds, env, pi, mu, loc -> (
      (* To evaluate a variable update with the write of a reference,
         we evaluate the first argument to a location, then we
         evaluate the second argument to a value, then we update the
         memory at the computed location with the value we obtained.
         Then, we add a binding in the environment to the unit value
         (this is the value produced by "set"), and we continue with
         the rest of the code. *)
      let vy = lookup_env env y in
      match vy with
      | Loc locy ->
          let vi = eval_imm env i in
          ( return_unit :: cmds,
            set_env env x Unit,
            pi,
            set_mem ~check_presence:true mu locy vi,
            loc )
      | _ -> raise (Error "Illegal SetRef: a location was expected"))
  | (While (cond, body_loop) as w) :: cmds, env, pi, mu, loc -> (
      (* To evaluate a while statement, simply unfold the loop once. *)
      match eval_imm env cond with
      | Bool b ->
          ( (if b then body_loop @ (w :: cmds) else return_unit :: cmds),
            env,
            pi,
            mu,
            loc )
      | _ -> raise (Error "Illegal 'while': a boolean was expected"))
  | [ Return (Call (f, args)) ], env, pi, mu, loc when tailcalls ->
      (* To evaluate a function call in a terminal position, we first look
         up the function (i.e. the formal argument and the code) that is
         associated to the function name [f]. There must be as many formal
         arguments as there are actual argument at the call site (otherwise
         we fail). We evaluate each of the arguments, and build an
         environment that associates to the formal arguments the actual
         values that we have just computed at this call site. This
         environment is set as the new environment in which we run the
         function's code. This is a tail call (the code tells us that
         nothing else must be computed next), so the call stack is left
         unchanged. Remark: the environment of the caller is completely
         discarded. *)
      let f_args, code = lookup_fun program f in
      if List.compare_lengths f_args args <> 0 then
        raise
          (Error
             ("Formal and actual number of parameters do not match in a call \
               to " ^ f));
      let vargs = List.map (eval_imm env) args in
      let init_env = List.combine f_args vargs in
      (code, init_env, pi, mu, loc)
  | Return (Call (f, args)) :: cmds, env, pi, mu, loc ->
      (* To evaluate a function call in a non-terminal position, we retrieve
         the formal arguments and the code for the function to be called,
         and we check that there are as many actual arguments provided at
         this call site as is required by the list of formal arguments. Then
         we evaluate the actual arguments and create an initial environment
         to execute the function, that binds the formal parameters to the
         values of the actual parameters. We set the code to execute to the
         body of the function, and add on top of the call stack the
         information of: what is the code to execute after the function has
         returned, and in which environment it should be executed. *)
      let f_args, code = lookup_fun program f in
      if List.compare_lengths f_args args <> 0 then
        raise
          (Error
             ("Formal and actual number of parameters do not match in a call \
               to " ^ f));
      let vargs = List.map (eval_imm env) args in
      let init_env = List.combine f_args vargs in
      (code, init_env, (None, cmds, env) :: pi, mu, loc)
  | AssignVar (x, Call (f, args)) :: cmds, env, pi, mu, loc ->
      (* To evaluate a variable update with a function call in a
         non-terminal position, we retrieve the formal arguments and the
         code for the function to be called, and we check that there are as
         many actual arguments provided at this call site as is required by
         the list of formal arguments. Then we evaluate the actual arguments
         and create an initial environment to execute the function, that
         binds the formal parameters to the values of the actual parameters.
         We set the code to execute to the body of the function, and add on
         top of the call stack the information of: where the value produced
         by the function should be written, what is the code to execute
         after the function has returned, and in which environment it should
         be executed. *)
      let f_args, code = lookup_fun program f in
      if List.compare_lengths f_args args <> 0 then
        raise
          (Error
             ("Formal and actual number of parameters do not match in a call \
               to " ^ f));
      let vargs = List.map (eval_imm env) args in
      let init_env = List.combine f_args vargs in
      (code, init_env, (Some x, cmds, env) :: pi, mu, loc)

(** [eval ~trace ~interactive ~tailcalls ~main p] evaluates the
   program [p] by calling the function named "main" on zero arguments.
   If [trace = true] or [interactive = true], then the configurations
   of each step of the evaluation are printed. Moreover, when
   [interactive = true], the evaluator waits for a user input between
   each evaluation step. If [tailcalls = true], then calls in tail
   positions do not consume an extra stack frame. *)
let eval ~trace ~interactive ~tailcalls ~main (p : program) : (v, string) result
    =
  let trace = trace || interactive in
  let open PPrint in
  Format.printf "Main function: %a@." pp_fname main;
  if trace then
    Format.printf "Execution trace of the program:@.@[%a@]@." pp_program p;
  let max_stack_size = ref 0 and nb_steps = ref 0 in
  let init_conf = ([ Return (Call (main, [])) ], [], [], [], 0) in
  let rec steps conf =
    if trace then
      Format.printf "Configuration:@.@[%a@]@.@." pp_configuration conf;
    if interactive then (
      print_endline "Hit 'Enter' to proceed with the next evaluation step...";
      ignore (read_line ()));
    max_stack_size := max !max_stack_size (stack_size conf);
    match is_final conf with
    | Some v -> v
    | None ->
        let conf' = step ~tailcalls p conf in
        incr nb_steps;
        steps conf'
  in
  match steps init_conf with
  | v ->
      Format.printf "@[Result:@ @[%a@]@]@." pp_v v;
      Format.printf "Maximum stack size: %i@." !max_stack_size;
      Format.printf "Number of evaluation steps: %i@.@." !nb_steps;
      Ok v
  | exception Error msg -> Error msg
